private import std.conv;
private import std.string;
private import std.date;
private import std.outbuffer;
private import std.c.stdlib;

private import std.socket;
private import sockutil;

private import myio;
private import log;
private import bogoyaml;
private import dutil;
private import request;
private import response;
private import responser;
private import typical_responses;
private import std_responser;
private import filter;
private import client;

version (Win32) {
	private import create_child;
	private import win32.ansi.winbase;
}

version (USE_CGI_RESPONSER) {
	private import cgi_responser;
}

version (USE_YAML_RESPONSER) {
	private import yaml_responser;
	private import yaml_filter;
}

version (USE_RC_RESPONSER) {
	private import rc_responser;
	private import rc_filter;
}

version (USE_DIRECTORY_RESPONSER) {
	private import directory_responser;
}

version (USE_PROXY_RESPONSER) {
	private import proxy_responser;
}

version (USE_SYS_RESPONSER) {
    private import sys_responser;
}

version (USE_RESPONSE_FILTER) {
	private import response_filter;
}

version (USE_GZIP_FILTER) {
	private import gzip_filter;
}

version (USE_CACHE_RESPONSER) {
	private import cache_responser;
}

extern (C) {
	AttoHttpd httpd;
    version (linux) {
        extern (C) void* signal(int sig, void* p);
    }

	//int cmain(char* arg) {
	int cmain() {
		//httpd = new AttoHttpd(arg);
		httpd = new AttoHttpd();
        version (linux) {
            version (USE_SYS_RESPONSER) {
                signal(2, cast(void*)&kill_attohttpd); // SIGINT
            }
        }
        int ret = httpd.run();
        quit();
		return ret;
	}

    void quit() {
        AttoHttpd h = httpd;
        httpd = null;
        h.quit();
    }

	bool AttoHttpd_Running() {
        if (httpd === null) return false;
		return httpd.running();
	}
}

class AttoHttpd {
private:
	static AttoHttpd self_;

public:
	static AttoHttpd self() { return self_; }

public:
	//this(char* arg) {
	this() {
		self_ = this;

		//initConf(arg);
		initConf();
		Log.info("config file loaded");

		initMisc();

		if (conf_.get("DenyExternal").integer() != 0) {
			responsers_ ~= new DenyExternalResponser();
		}

		responsers_ ~= new LocalFileResponser();

		version (USE_SYS_RESPONSER) {
			if (conf_.get("EnableSysResponser").integer) {
				responsers_ ~= new KillResponser();
			}
		}

		version (USE_CACHE_RESPONSER) {
			if (conf_.get("EnableCacheResponser").integer) {
				responsers_ ~= new CacheResponser();
			}
		}

		version (USE_YAML_RESPONSER) {
			if (conf_.get("EnableYamlResponser").integer) {
				responsers_ ~= new YamlResponser();
				filters_ ~= new YamlFilter();
			}
		}

		version (USE_RC_RESPONSER) {
			if (conf_.get("EnableRcResponser").integer) {
				char[] sep = conf_.get("RcSeparator").str;
				responsers_ ~= new RcResponser(sep);
				filters_ ~= new RcFilter(sep);
			}
		}

		version (USE_CGI_RESPONSER) {
			if (conf_.get("EnableCgiResponser").integer) {
				responsers_ ~= new CgiResponser();
			}
		}

		version (USE_PROXY_RESPONSER) {
			if (conf_.get("EnableProxyResponser").integer) {
				responsers_ ~= new ProxyResponser();
			}
		}

		version (USE_GZIP_FILTER) {
			if (conf_.get("EnableGzipFilter").integer) {
				filters_ ~= new GzipFilter();
			}
		}

		version (USE_RESPONSE_FILTER) {
			if (conf_.get("EnableResponseFilter").integer) {
				filters_ ~= new ResponseFilter();
			}
		}

		version (USE_CACHE_RESPONSER) {
			if (conf_.get("EnableCacheResponser").integer) {
				filters_ ~= new CacheFilter();
			}
		}

		responsers_ ~= new NormalResponser();

		version (USE_DIRECTORY_RESPONSER) {
			responsers_ ~= new DirectoryResponser();
		}

		responsers_ ~= new FinalResponser();
	}

	this(Responser[] responsers) {
		this(responsers, cast(char*)0);
	}

	this(Responser[] responsers, char* arg) {
		self_ = this;

		//initConf(arg);
		initConf();
		Log.info("config file loaded");

		responsers_ = responsers;
	}

	~this() {
        quit();
	}

    void quit() {
		if (serverSocket_ !== null) {
            Log.info("close server socket");
			serverSocket_.close();
			serverSocket_ = null;
		}
		if (socketset_ !== null) {
			socketset_ = null;
		}

		Log.info("attohttpd end");
    }

	int run() {
		initNet();

		Log.info("wait a socket using port " ~ tostr(port_));

		running_ = true;
		firstTime_ = std.date.getUTCtime();

		procBrowserChain();

		while (running_) {
			uint elapsed =
				(std.date.getUTCtime()-firstTime_) / std.date.TicksPerSecond;
			int timeout = (uptime_ - elapsed) * 1000000;

			if (uptime_ <= elapsed) {
				running_ = false;
				Log.info(tostr(elapsed) ~
						 " seconds elapsed, attohttpd shutdown.");
				break;
			}

            int ret = 0;
            if (uptime_ == -1) ret = Socket.select(socketset_, null, null);
            else ret = Socket.select(socketset_, null, null);

            if (ret) {
				handleServer();
			}

            for (int i = 0; i < clients_.length; i++) {
                if (clients_[i].isTerminated()) {
                    clients_ = clients_[0..i] ~ clients_[i+1..clients_.length];
                    i--;
                }
            }
	  	}

 		running_ = false;

        foreach (Client c; clients_) {
            if (!c.isTerminated()) {
                c.wait(5000);
                if (!c.isTerminated()) {
                    Log.warn("ignore running thread");
                }
            }
        }

		return 0;
	}

	bool running() {
		return running_;
	}
	void end() {
		running_ = false;
	}

	YamlHolder conf() { return conf_; }

    InternetAddress address() { return serverIp_; }

	int port() { return port_; }
	bool verbose() { return verbose_; }

    Responser[] responsers() { return responsers_; }

private:
	//void initConf(char* arg) {
	void initConf() {
		try {
			//if (arg) {
			//	conf_ = yamlLoad(std.string.toString(arg));
			//}
			//else {
            conf_ = yamlLoad("attohttpd.yml");
			//}
	    }
		catch (Exception e) {
			Log.error(e.toString());
		}

		port_ = conf_.get("Port").integer();

		uptime_ = conf_.get("Uptime").integer();

        verbose_ = conf_.get("Verbose").integer() != 0;
	}

	void initMisc() {
	}

	void initNet() {
        int shifts = conf_.get("AutoShiftNextPort").integer();
        for (int i = 0; i < shifts; i++) {
            try {
                serverIp_ = new InternetAddress(port_);

                serverSocket_ = new Socket(AddressFamily.INET,
                                           SocketType.STREAM,
                                           ProtocolType.IP);

                socketset_ = new SocketSet(1);
                socketset_.add(serverSocket_);
                //serverSocket_.blocking = false;

                version (linux) {
                    const int SOL_SOCKET = 1;
                }
                version (Win32) {
                    const int SOL_SOCKET = 0xffff;
                }

                int one = 1;
                serverSocket_.setOption(cast(SocketOptionLevel)SOL_SOCKET,
                                        SocketOption.REUSEADDR, one);
/*
				serverSocket_.setOption(SocketOptionLevel.SOCKET,
                                        SocketOption.REUSEADDR, one);
*/

                serverSocket_.bind(serverIp_);
                serverSocket_.listen(10);

                return;
            }
            catch (SocketException se) {
                Log.warn("cannot create server socket with port " ~
                         tostr(port_) ~ ": " ~ se.toString());
                port_++;
            }
        }
        Log.error("cannot create server socket");
	}

	void handleServer() {
        Socket sock;
        try {
            sock = serverSocket_.accept();
        }
        catch (SocketException se) {
			Log.warn("cannot accept socket: " ~ se.toString());
			return;
		}

		Address ip = sock.remoteAddress();
		Log.info("connection established with " ~ ip.toString());

        Client c = new Client(sock, responsers_, filters_);
        clients_ ~= c;
        c.start();
	}

	void procBrowserChain() {
		version (Win32) {
			try {
				if (conf_.get("WindowsBrowserChain").get("Enable").integer == 0)
					return;

				char[] browser =
					conf_.get("WindowsBrowserChain").get("Browser").str;
				char[] file = conf_.get("WindowsBrowserChain").get("File").str;

				char[] cmd =
					browser ~ " http://localhost:" ~ tostr(port_) ~ "/" ~ file;

				win32.ansi.winbase.SECURITY_ATTRIBUTES sa;
				sa.nLength = (win32.ansi.winbase.SECURITY_ATTRIBUTES).sizeof;
				sa.lpSecurityDescriptor = NULL;
				sa.bInheritHandle = win32.ansi.windef.TRUE;

				Log.info("execute " ~ cmd);

				HANDLE child = CreateChild(toStringz(cmd), NULL, &sa,
										   NULL, NULL, NULL);
			}
			catch (Error e) {
				Log.warn(e.toString());
			}
		}
	}

private:
	YamlHolder conf_;

	SocketSet socketset_;
	Socket serverSocket_;
    Client[] clients_;
	InternetAddress serverIp_;

private:
	bool running_;

private:
	uint port_;

	d_time firstTime_;
	uint uptime_;
    bool verbose_;

private:
	Responser[] responsers_;
	Filter[] filters_;

}
