version (Win32) {
	private import std.string;

	private import win32.ansi.winbase;

	alias int pid_t;

	static HANDLE
		CreateChild(char *cmd, char *prog, SECURITY_ATTRIBUTES *psa,
					HANDLE hInput, HANDLE hOutput, HANDLE hError) {
		BOOL fRet;
		DWORD  dwCreationFlags;
		STARTUPINFO aStartupInfo;
		PROCESS_INFORMATION aProcessInformation;
		SECURITY_ATTRIBUTES sa;
		char *shell;
		HANDLE child;

		if (!cmd && !prog) {
			return NULL;
		}

		if (!psa) {
			sa.nLength              = SECURITY_ATTRIBUTES.sizeof;
			sa.lpSecurityDescriptor = NULL;
			sa.bInheritHandle       = win32.ansi.windef.TRUE;
			psa = &sa;
		}

		std.string.memset(&aStartupInfo, 0, STARTUPINFO.sizeof);
		std.string.memset(&aProcessInformation, 0, PROCESS_INFORMATION.sizeof);
		aStartupInfo.cb = STARTUPINFO.sizeof;

		if (hInput || hOutput || hError) {
			aStartupInfo.dwFlags = STARTF_USESTDHANDLES;
			if (hInput) {
				aStartupInfo.hStdInput  = hInput;
			}
			else {
				aStartupInfo.hStdInput  = win32.ansi.winbase.GetStdHandle(STD_INPUT_HANDLE);
			}
			if (hOutput) {
				aStartupInfo.hStdOutput = hOutput;
			}
			else {
				aStartupInfo.hStdOutput = win32.ansi.winbase.GetStdHandle(STD_OUTPUT_HANDLE);
			}
			if (hError) {
				aStartupInfo.hStdError = hError;
			}
			else {
				aStartupInfo.hStdError = win32.ansi.winbase.GetStdHandle(STD_ERROR_HANDLE);
			}
		}

		dwCreationFlags = (NORMAL_PRIORITY_CLASS);

		fRet = CreateProcess(NULL, cmd, psa, psa,
							 psa.bInheritHandle, dwCreationFlags, NULL, NULL,
							 &aStartupInfo, &aProcessInformation);

		if (!fRet) {
			return NULL;
		}

		CloseHandle(aProcessInformation.hThread);

		child = aProcessInformation.hProcess;

		/+
			if (!IsWinNT()) {
				/* On Win9x, make pid positive similarly to cygwin and perl */
				child.pid = -child.pid;
			}
		+/

		return child;
	}
}
