// why phobos'es std.zlib.uncompress cannot inflate gzip?

private import std.zlib;
private import etc.c.zlib;

static int ASCII_FLAG = 0x01; /* bit 0 set: file probably ascii text */
static int HEAD_CRC = 0x02; /* bit 1 set: header CRC present */
static int EXTRA_FIELD = 0x04; /* bit 2 set: extra field present */
static int ORIG_NAME = 0x08; /* bit 3 set: original file name present */
static int COMMENT = 0x10; /* bit 4 set: file comment present */
static int RESERVED = 0xE0; /* bits 5..7: reserved */

int checkHeader(ubyte[] buf) {
    int i = 2;
    int method = buf[i++];
    int flags  = buf[i++];

    if (method != Z_DEFLATED || (flags & RESERVED) != 0) {
        return -1;
    }

    i+=6;

    if (flags & EXTRA_FIELD) { /* skip the extra field */
        int len = 0;
        len  =  cast(uint)buf[i++];
        len += (cast(uint)buf[i++])<<8;
        /* len is garbage if EOF but the loop below will quit anyway */
        i+=len;
    }
    if (flags & ORIG_NAME) { /* skip the original file name */
        while (buf[i++] !=0) {}
    }
    if (flags & COMMENT) {   /* skip the .gz file comment */
        while (buf[i++] != 0) {}
    }
    if (flags & HEAD_CRC) {  /* skip the header crc */
        i+=2;
    }
    return i;
}

void[] uncompress(void[] s) {
    ubyte[] gis = cast(ubyte[])s;
    ubyte[1024] buf;
    ubyte[] gos;

    int h = checkHeader(gis);

    gis = gis[h..gis.length];

    z_stream z;
    z.zalloc = cast(alloc_func)Z_NULL;
    z.zfree = cast(free_func)Z_NULL;
    z.opaque = cast(void*)Z_NULL;

    int status = inflateInit2(&z, -15);
    if (status != Z_OK) {
        throw new ZlibException(status);
    }

    z.next_in = gis;
    z.avail_in = gis.length;

    z.next_out = buf;
    z.avail_out = buf.length;

    status = Z_OK;
    while (status != Z_STREAM_END) {
        if (z.avail_in == 0) {
            break;
        }

        status = inflate(&z, std.zlib.Z_NO_FLUSH);
        if (status == Z_STREAM_END) {
            int used = buf.length - z.avail_out;
            gos ~= buf[0..used];
            break;
        }
        if (status != Z_OK) {
            throw new ZlibException(status);
            break;
        }
        if (z.avail_out == 0) {
            gos ~= buf;
            z.next_out = buf;
            z.avail_out = buf.length;
        }
    }

    return gos;
}

