private import std.stream;
private import std.string;
private import std.c.stdio;
private import std.c.stdlib;

private import create_child;

// please remove it if you are using this file for other project
version (USE_CGI_RESPONSER) {
	version = USE_PIPE;
}
version (USE_PIPE) {

enum PipeMode {
	None = 0,
	In = 1,
	Out = 2,
	InOut = 3
}

class StreamPipeError: StreamError {
	this(char[] msg) { super(msg); }
}

version (linux) {
	private import std.c.linux.linux;

	extern(C) int pipe(int filedes[2]);
	extern(C) int dup2(int oldfd, int newfd);
	extern(C) char *strerror(int errnum);
	alias int pid_t;
	extern(C) pid_t fork();
	extern(C) int system(char *string);
	extern(C) int execv(char *path, char*[] argv);
	extern(C) pid_t waitpid(pid_t pid, int *status, int options);

	alias std.c.linux.linux linux;
}
version (Win32) {
	private import win32.ansi.winbase;
}

class Pipe : Stream {
public:
	override ulong seek(long offset, SeekPos rel) {
//		throw new SeekError("pipe cannot seek");
		return 0;
	}

	bool canRead() { return (mode_ & PipeMode.In) != 0; }
	bool canWrite() { return (mode_ & PipeMode.Out) != 0; }

private:
	PipeMode mode_;

public:
	this(char[] cmd, PipeMode mode) {
		this(cmd, mode, null);
	}

	this(char[] cmd, PipeMode mode, char[][] argv) {
		if (mode != PipeMode.In &&
			mode != PipeMode.Out &&
			mode != PipeMode.InOut)
		{
			throw new StreamPipeError("invalid mode");
		}
		mode_ = mode;

		version (linux) {
			if (canRead() && pipe(readfd) < 0) {
				throwError();
			}
			if (canWrite() && pipe(writefd) < 0) {
				throwError();
			}

			if ((pid = fork()) == 0) {
				if (canWrite()) {
					linux.close(writefd[1]);
					dup2(writefd[0], 0);
					linux.close(writefd[0]);
				}
				if (canRead()) {
					linux.close(readfd[0]);
					dup2(readfd[1], 1);
					linux.close(readfd[1]);
				}

				char* c = toStringz(cmd);

/*
  char*[] argvz;
  argvz ~= c;
  if (argv !== null) {
  foreach (char[] a; argv) {
  argvz ~= string.toStringz(a);
  }
  }
  argvz ~= (char*)0;

  execv(c, argvz);
*/
				system(c);

				if (canWrite()) {
					linux.close(writefd[0]);
				}

				exit(0);
			}

			if (canRead()) linux.close(readfd[1]);
			if (canWrite()) linux.close(writefd[0]);
		}
		version (Win32) {
		    HANDLE hCurProc = GetCurrentProcess();
			HANDLE hReadIn, hReadOut;
			HANDLE hWriteIn, hWriteOut;
			HANDLE hDupInFile, hDupOutFile;

			SECURITY_ATTRIBUTES sa;
			sa.nLength = SECURITY_ATTRIBUTES.sizeof;
			sa.lpSecurityDescriptor = NULL;
			sa.bInheritHandle = win32.ansi.windef.TRUE;

			if (canRead()) {
				CreatePipe(&hReadIn, &hReadOut, &sa, 2048L);
				DuplicateHandle(hCurProc, hReadIn, hCurProc, &hDupInFile,
								0, FALSE, DUPLICATE_SAME_ACCESS);
				CloseHandle(hReadIn);
			}
			if (canWrite()) {
				CreatePipe(&hWriteIn, &hWriteOut, &sa, 2048L);
				DuplicateHandle(hCurProc, hWriteIn, hCurProc, &hDupOutFile,
								0, FALSE, DUPLICATE_SAME_ACCESS);
				CloseHandle(hWriteOut);
			}
			CloseHandle(hCurProc);

		    HANDLE child = CreateChild(toStringz(cmd), NULL, &sa,
									   hWriteIn, hReadOut, NULL);

			if (!child) {
				if (canRead()) {
					CloseHandle(hReadOut);
					CloseHandle(hDupInFile);
				}
				if (canWrite()) {
					CloseHandle(hWriteIn);
					CloseHandle(hDupOutFile);
				}
				throw new StreamPipeError(cmd ~ ": cannot exec");
			}

			if (canRead()) {
				CloseHandle(hReadOut);
			}
			if (canWrite()) {
				CloseHandle(hWriteIn);
			}

			hIn = hDupInFile;
			hOut = hDupOutFile;
				
		}
	}

	~this() {
		close();
	}

	void throwError() {
		version (linux) {
			throw new StreamPipeError(
				std.string.toString(strerror(getErrno())));
		}
		version (Win32) {
			throw new StreamPipeError("pipe errored");
		}
	}

	override void close() {
		version (linux) {
			if (canRead()) {
				linux.close(readfd[0]);
			}
			if (canWrite()) {
				linux.close(writefd[1]);
			}
			waitpid(pid, cast(int*)0, 1);
			mode_ = PipeMode.None;
		}
		version (Win32) {
			if (canRead()) {
				CloseHandle(hIn);
			}
			if (canWrite()) {
				CloseHandle(hOut);
			}
			CloseHandle(child);
			mode_ = PipeMode.None;
		}
	}

	override uint readBlock(void* buffer, uint size) {
		if (!canRead()) {
			throw new StreamPipeError("pipe is not readable");
		}
		version (linux) {
			int ret = linux.read(readfd[0], buffer, size);
//		fprintf(std.c.stdio.stderr, "%d\n", ret);
			return ret;
		}
		version (Win32) {
			uint s;
			ReadFile(hIn, buffer, size, &s, NULL);
			return s;
		}
	}

	override uint writeBlock(void* buffer, uint size) {
		if (!canWrite()) {
			throw new StreamPipeError("pipe is not writable");
		}
		version (linux) {
			return linux.write(writefd[1], buffer, size);
		}
		version (Win32) {
			uint s;
			WriteFile(hOut, buffer, size, &s, NULL);
			return s;
		}
	}

private:
	version (linux) {
		int readfd[2];
		int writefd[2];
		pid_t pid;
	}
	version (Win32) {
		HANDLE child;
		HANDLE hIn;
		HANDLE hOut;
	}
		
}


version (linux) {
private:
	extern (C) int setenv(char *name, char *value, int overwrite);
}

unittest {
	version (linux) {
		/*
		  {
		  static char[][] argv = [ "hoge" ];
		  auto Pipe p = new Pipe("./out.pl", PipeMode.In);
		  assert(!p.canWrite());
		  assert(p.canRead());
		  auto char[] str = p.readString(5);
		  std.stream.stdout.writeLine(str);
		  assert(str == "hoge\n");
		  }
		  {
		  auto Pipe p = new Pipe("./log.pl", PipeMode.Out);
		  assert(p.canWrite());
		  assert(!p.canRead());
		  p.writeLine("output test\n");
		  }
		  {
		  auto Pipe p = new Pipe("/bin/echo", PipeMode.In);
		  assert(!p.canWrite());
		  assert(p.canRead());
		  char[] buf;
		  buf.length = 1024;
		  buf.length = p.readBlock(buf, buf.length);
		  assert(buf == "\n");
		  }
		  {
		  auto Pipe p = new Pipe("/bin/echo", PipeMode.In);
		  assert(!p.canWrite());
		  assert(p.canRead());
		  char[] buf;
		  buf.length = 1024;
		  buf.length = p.readBlock(buf, buf.length);
		  assert(buf == "\n");
		  }
		*/
		{
			setenv("CONTENT_LENGTH", "9", 1);
			setenv("REQUEST_METHOD", "POST", 1);

			auto Pipe p = new Pipe("perl ../cgi/env.cgi", PipeMode.InOut);
			assert(p.canWrite());
			assert(p.canRead());
			p.writeLine("hoge=huga\n");

			char[] all = "";
			char c;
			try {
				while ((c = p.getc()) != '\0') {
					all ~= c;
//				dout.writeLine(all);
				}
			}
			catch (Object o) {}
//			assert(str == "hoge\n");
//			std.stream.stdout.writeLine(all);
		}

	}
}


}


