private import responser;
private import typical_responses;
private import response;
private import request;
private import log;
private import attohttpd;
private import sockutil;
private import myio;

private import std.socket;

private import std.c.stdio;
private import std.c.stdlib;
private import std.c.time;
private import std.stream;
private import std.string;
private import std.conv;
private import std.ctype;
private import std.thread;

version (USE_PROXY_RESPONSER) {
    private static HttpResponse error(char[] msg) {
        Log.warn(msg);
        return TypicalResponses.resProxyError(msg);
    }

    class ProxyThread : Thread {
        HttpRequest req;
        HttpResponse res;
        char[] state;

        this(HttpRequest r) {
            req = r;
            state = "not started";
        }

        override int run() {
            Socket sock;

            with (req) {
                if (!requestHost) return 0;

                state = "resolve name";
                int port;
                InternetAddress server;
                try {
                    port = (requestPort) ? toInt(requestPort) : 80;
                    server = new InternetAddress(requestHost, port);
                }
                catch (SocketException se) {
                    res = .error(requestHost ~ ":" ~ tostr(port) ~
                                 ": cannot resolve hostname: " ~ se.toString());
                    return 0;
                }

                if ((server.addr == AttoHttpd.self.address.addr ||
                     server.addr == LOCALHOST) &&
                    server.port == AttoHttpd.self.address.port)
                {
                    return 0;
                }

                state = "send request";
                try {
                    sock = new Socket(AddressFamily.INET, SocketType.STREAM,
                                      ProtocolType.IP);
                    sock.connect(server);
                }
                catch (SocketException se) {
                    res = .error(requestHost ~ ":" ~ tostr(port) ~
                                 ": connection failed: " ~ se.toString());
                    return 0;
                }

                SocketSet sockset;
                try {
                    sockset = new SocketSet(1);
                }
                catch (SocketException se) {
                    res =  .error("cannot create socket set: " ~ se.toString());
                    return 0;
                }
                sockset.add(sock);

                safeSend(sock, all);
            }

            state = "receive response";

            static const int BUFSIZE = 4 * 1024;

            char[] data;
            char[] buf;
            buf.length = BUFSIZE;
            int allLen = 0;
            for (;;) {
                int len = sock.receive(buf, SocketFlags.NONE);

                if (len == 0) {
                    break;
                }
                else {
                    data ~= buf;
                }
            }

            res = HttpResponse.createFromFullResponse(tostr(data));

            return 0;
        }
    }

    class ProxyResponser : Responser {
    public:
        this() {
            timeout_ =
                AttoHttpd.self.conf.get("ProxyTimeout").integer;
        }

        override HttpResponse response(HttpRequest req) {
            ProxyThread thread = new ProxyThread(req);
            thread.start();

            int times = timeout_ * 10;
            for (int i = 0; i < times; i++) {
                usleep(100000);
                if (thread.getState() == Thread.TS.TERMINATED) {
                    return thread.res;
                }
            }

            return error("proxy request timeouted (" ~ thread.state ~ ")");
        }

/+
            Socket sock;

            with (req) {
                if (!requestHost) return null;

                int port;
                InternetAddress server;
                try {
                    port = (requestPort) ? toInt(requestPort) : 80;
                    server = new InternetAddress(requestHost, port);
                }
                catch (SocketException se) {
                    return error(requestHost ~ ":" ~ tostr(port) ~
                                 ": cannot resolve hostname: " ~ se.toString());
                }

                if ((server.addr == AttoHttpd.self.address.addr ||
                     server.addr == LOCALHOST) &&
                    server.port == AttoHttpd.self.address.port)
                {
                    return null;
                }

                try {
                    sock = new Socket(AddressFamily.INET, SocketType.STREAM,
                                      ProtocolType.IP);
                    sock.connect(server);
                }
                catch (SocketException se) {
                    return error(requestHost ~ ":" ~ tostr(port) ~
                                 ": connection failed: " ~ se.toString());
                }

                SocketSet sockset;
                try {
                    sockset = new SocketSet(1);
                }
                catch (SocketException se) {
                    return error("cannot create socket set: " ~ se.toString());
                }
                sockset.add(sock);

                safeSend(sock, all);
            }

            SocketStream data = new SocketStream(sock);

            HttpResponse res = new HttpResponse();

            with (res) {
                statusLine = data.readLine();

                int index = 0;
                while (index < statusLine.length &&
                       !isspace(statusLine[index]))
                    index++;

                // it's HTTP/0.9
                if (index < 4 || statusLine[0..4] != "HTTP") {
                    messageBody = statusLine ~ "\n";
                    statusLine = null;
                    try {
                        char c;
                        while ((c = data.getc()) != '\0') {
                            messageBody ~= c;
                        }
                    }
                    catch (ReadError re) {}

                    return res;
                }

                char[] httpVersion = statusLine[0..index];

                while (isspace(statusLine[index])) index++;
                int index2 = index;

                while (!isspace(statusLine[index])) index++;
                char[] code = statusLine[index2..index];

                while (isspace(statusLine[index])) index++;
                char[] message = statusLine[index..statusLine.length];

                int len = -1;
                bool transfer_encoding_chunked = false;
                char[] line = data.readLine();
                while (line != "") {
                    int f = find(line, ": ");
                    char[] k = line[0..f];
                    char[] v = line[f+2..line.length];

                    char[] lk = std.string.tolower(k);
                    if (lk == `content-length`) {
                        len = toInt(v);
                    }
                    else if (lk == `transfer-encoding`) {
                        if (std.string.tolower(v) == `chunked`) {
                            transfer_encoding_chunked = true;
                        }
                    }

                    header[k] = v;

                    line = data.readLine();
                }
                if (transfer_encoding_chunked) {
                    messageBody = "";
                    char[] lenStr = data.readLine();
                    sscanf(toStringz(lenStr), "%x", &len);
                    while (len) {
                        messageBody ~= data.readString(len);
                        lenStr = data.readLine();
                        lenStr = data.readLine();
                        sscanf(toStringz(lenStr), "%x", &len);
                    }
                    lenStr = data.readLine();
                }
                else if (len == -1) {
                    static const int BUFSIZE = 1024;
                    ubyte[BUFSIZE] buf;
                    int l = data.read(buf);
                    messageBody = "";
                    messageBody ~= cast(char[])buf[0..l];
                    while (l == BUFSIZE) {
                        l = data.read(buf);
                        messageBody ~= cast(char[])buf[0..l];
                    }
                }
                else {
                    messageBody = data.readString(len);
                }
            }

            return res;

/*
  static const int BUFSIZE = 1024 * 1024;
  static char buf[BUFSIZE];
  char[] data = "";

  for (;;) {
  if (Socket.select(sockset, null, null, timeout_)) {
  int len = sock.receive(buf);
  if (len <= 0) {
  break;
  }
  else {
  data ~= buf[0..len];
  Log.info(tostr(len));
  }
  }
  else {
  sock.shutdown(SocketShutdown.BOTH);
  sock.close();
  sockset.reset();
  return error("proxy request timeouted");
  }
  }
            
  sock.shutdown(SocketShutdown.BOTH);
  sock.close();

  return HttpResponse.createFromFullResponse(data);
*/
+/
            private int timeout_;
    }

}


