private import std.string;

private import myio;

class HttpRequest {
public:
	this() {}
	this(char[] data) {
		all = data;

		statusLine = nextLine(data);
		char[][] elems = std.string.split(statusLine, " ");
		if (elems.length > 0) method = elems[0];
		if (elems.length > 1) url = elems[1];
		if (elems.length > 2) httpVersion = elems[2];
		
		for (char[] line = nextLine(data); line != ""; line = nextLine(data)) {
			int idx = std.string.find(line, ":");

			// rough.
			if (idx == -1) continue;

			int sidx = idx + 1;
			while (line.length > sidx && line[sidx] == ' ') sidx++;

			char[] key = std.string.tolower(line[0..idx]);
			header[key] = line[sidx..line.length];
		}

		messageBody = data;
		if (messageBody.length >= 2 &&
			messageBody[messageBody.length-2..messageBody.length] == "\r\n")
		{
			messageBody.length = messageBody.length - 2;
		}
	}

    char[] toString() { return all; }

private:
	char[] nextLine(inout char[] data) {
		int idx = std.string.find(data, "\r\n");
		char[] ret;
		if (idx == -1) {
			ret = data;
			data = "";
		}
		else {
			ret = data[0..idx];
			data = data[idx+2..data.length];
		}
		return ret;
	}

public:
	char[] all;

	char[] statusLine;
	char[] method;
	char[] url;
	char[] httpVersion;
	char[][char[]] header;
	char[] messageBody;

	char[] requestHost;
	char[] requestPort;
	char[] localFile;
	char[] queryString;
	char[] dirName;
	char[] baseName;
	char[] suffix;

	uint host;
	int port;

    bool isLocal;
}

unittest {
	char[] r;
	HttpRequest req;

	r = "GET / HTTP/1.0\r\n" ~
		"User-Agent: hoge\r\n";
	req = new HttpRequest(r);
	assert(req.all == r);
	assert(req.method == "GET");
	assert(req.url == "/");
	assert(req.httpVersion == "HTTP/1.0");
	assert(req.header["user-agent"] == "hoge");
	assert(req.messageBody == "");
	assert(req.localFile == "");
	assert(req.queryString == null);

	r = "POST http://localhost/index.html HTTP/1.1\r\n" ~ 
		"Host: localhost\r\n" ~
		"Content-Length: 10\r\n" ~
		"\r\n" ~
		"0123456789\r\n";
	req = new HttpRequest(r);
	assert(req.all == r);
	assert(req.method == "POST");
	assert(req.url == "http://localhost/index.html");
	assert(req.httpVersion == "HTTP/1.1");
	assert(req.header["content-length"] == "10");
	assert(req.header["host"] == "localhost");
	assert(req.messageBody == "0123456789");
//	assert(req.localFile == "index.html");
//	assert(req.queryString == null);

	r = "GET /hoge.cgi?q=fuga";
	req = new HttpRequest(r);
	assert(req.all == r);
	assert(req.method == "GET");
	assert(req.url == "/hoge.cgi?q=fuga");
	assert(req.httpVersion === null);
	assert(req.header.length == 0);
	assert(req.messageBody == "");
//	assert(req.localFile == "hoge.cgi");
//	assert(req.queryString == "q=fuga");
}
