private import std.file;
private import std.string;
private import std.stream;
private import std.path;
private import std.conv;

private import request;
private import response;
private import log;
private import typical_responses;
private import bogoyaml;
private import fileutil;
private import responser;
private import attohttpd;
private import sockutil;
private import myio;

class DenyExternalResponser : Responser {
	HttpResponse response(HttpRequest req) {
		if (req.host != LOCALHOST) {
			return TypicalResponses.resForbidden;
		}
		return null;
	}
}

class LocalFileResponser : Responser {
	this() {
		YamlHolder conf = AttoHttpd.self.conf;
		indexes_ = conf.get("DirectoryIndex").array;
	}

	HttpResponse response(HttpRequest req) {
		with (req) {
			localFile = url.dup;
			if (url.length > 7 && url[0..7] == "http://") {
				localFile = url[7..url.length];
				int idx = std.string.find(localFile, "/");
				if (idx == -1) return null;
				requestHost = localFile[0..idx];
				localFile = localFile[idx..localFile.length];
				
				idx = std.string.find(requestHost, ":");
				if (idx != -1) {
					requestPort = requestHost[idx+1..requestHost.length];
					requestHost = requestHost[0..idx];
				}
			}

			if (localFile.length == 0 || localFile[0] != '/') return null;

			localFile = localFile[1..localFile.length];
			int idx = std.string.find(localFile, "?");
			if (idx != -1) {
				queryString = localFile[idx+1..localFile.length];
				localFile = localFile[0..idx];
			}

			if (localFile.length == 0) localFile = "./";

//			if (localFile.length == 0) return null;

			if (localFile[localFile.length-1] == '/') {
				foreach (YamlHolder index; indexes_) {
					if (file_canread(localFile ~ index.str)) {
						localFile ~= index.str;
					}
				}
			}

			suffix = std.path.getExt(localFile);
			dirName = std.path.getDirName(localFile);
			baseName = std.path.getBaseName(localFile);

            port = (requestPort) ? toInt(requestPort)
                : (requestHost) ? 80 : AttoHttpd.self.port;

            isLocal = (port == AttoHttpd.self.port &&
                       (requestHost === null ||
                        requestHost == "localhost" ||
                        requestHost == "127.0.0.1"));
		}

		return null;
	}

private:
	YamlHolder[] indexes_;
}

class NormalResponser : Responser {
	this() {
		YamlHolder conf = AttoHttpd.self.conf;
		types_ = conf.get("ContentType").assoc;
	}

	HttpResponse response(HttpRequest req) {
		if (!req.isLocal) {
			return null;
		}

		char[] localFile = req.localFile;

//		if (localFile.length == 0) return null;

		if (localFile[localFile.length-1] == '/') return null;

		if (!file_canread(localFile)) {
			// ignoring 403...
			return TypicalResponses.resNotFound;
		}

		HttpResponse res = new HttpResponse("HTTP/1.0 200 OK");

		char[] type = "text/plain";
		foreach (char[] s; types_.keys) {
			if (s == req.suffix) {
				 type = types_[s].str;
			}
		}

		res.header["Content-Type"] = type;

		res.messageBody = cast(char[])(std.file.read(localFile));

		res.addDefaultHeader();

		return res;
	}

private:
	YamlHolder[char[]] types_;
}

class FinalResponser : Responser {
public:
	HttpResponse response(HttpRequest req) {
		return TypicalResponses.resInternalServerError;
	}
}
